""" Functional class for SciANN.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python import keras as k

from keras import backend as K
graph_unique_name = K.get_graph().unique_name

from keras.layers import Dense
from keras.layers import Concatenate
from tensorflow.python.keras.layers import InputSpec
from keras.constraints import MinMaxNorm
from keras.constraints import Constraint
from tensorflow.python.util.tf_export import keras_export
from keras.utils import tf_utils
from keras import constraints
from keras import initializers
from keras import regularizers

from ..utils import to_list
from ..utils import default_constant_initializer
from ..utils import default_regularizer

from .functional import MLPFunctional
from .variable import Variable


class Parameter(MLPFunctional):
    """ Parameter functional to be used for parameter inversion.
        Inherited from Dense layer.

    # Arguments
        val: float.
            Initial value for the parameter.
        min_max: [MIN, MAX].
            A range to constrain the value of parameter.
            This constraint will overwrite non_neg constraint if both are chosen.
        inputs: Variables.
            List of `Variable`s to the parameters.
        name: str.
            A name for the Parameter layer.
        non_neg: boolean.
            True (default) if only non-negative values are expected.
        **kwargs: keras.layer.Dense accepted arguments.
        
    """
    def __init__(self,
                 val=1.0,
                 min_max=None,
                 inputs=None,
                 name=None,
                 non_neg=None):

        inputs = to_list(inputs)
        if not all([isinstance(x, (Variable, MLPFunctional)) for x in inputs]):
            raise TypeError

        input_tensors, layers = [], []
        for v in inputs:
            input_tensors += v.outputs
            layers += v.layers

        if non_neg is None:
            non_neg = True
        layers.append(
            ParameterBase(val=val, min_max=min_max, non_neg=non_neg, name=name)
        )

        if len(input_tensors) > 1:
            lay = Concatenate(name=graph_unique_name('conct'))
            lay_input = lay(input_tensors)
        else:
            lay_input = input_tensors[0]
        outputs = layers[-1](lay_input)

        super(Parameter, self).__init__(
            inputs=to_list(input_tensors),
            outputs=to_list(outputs),
            layers=to_list(layers)
        )

    @classmethod
    def get_class(cls):
        return MLPFunctional

    @property
    def name(self):
        return self.layers[-1].name

    @property
    def value(self):
        return K.get_value(self.layers[-1].param)


@keras_export('keras.layers.ParameterBase')
class ParameterBase(k.layers.Layer):
    """ Base Parameter class to be used for parameter inversion.
        Inherited from Dense layer.

    # Arguments
        val (float): Initial value for the parameter.
        min_max ([MIN, MAX, [Penalty]]): A range to constrain the value of parameter.
            This constraint will overwrite non_neg constraint if both are chosen.
        non_neg (boolean): True (default) if only non-negative values are expected.
        **kwargs: keras.layer.Dense accepted arguments.
    """

    def __init__(self, val=None, min_max=None, non_neg=True, **kwargs):
        super(ParameterBase, self).__init__(**kwargs)

        cst = None
        if min_max is not None:
            cst = MinMax(min_value=min_max[0], max_value=min_max[1], penalty=1.0 if len(min_max) == 2 else min_max[2])
            val = (min_max[0] + min_max[1]) / 2.0 if val is None else val
        elif non_neg:
            cst = k.constraints.non_neg()

        # Default value for initial values.
        val = 1.0 if val is None else val

        self.param_initializer = default_constant_initializer(val)
        self.param_regularizer = None
        self.param_constraint = None
        self.shared_axes = None
        self.param_constraint = cst

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        assert len(input_shape) >= 2
        input_dim = list(input_shape[1:])

        self.param = self.add_weight(
            shape=[1,],
            initializer=self.param_initializer,
            name='param',
            regularizer=self.param_regularizer,
            constraint=self.param_constraint
        )
        # Set input spec
        axes = {}
        if self.shared_axes:
            for i in range(1, len(input_shape)):
                if i not in self.shared_axes:
                    axes[i] = input_shape[i]
        self.input_spec = InputSpec(ndim=len(input_shape), axes=axes)
        self.built = True

    def call(self, inputs):
        return self.param #+ inputs*0.0

    def get_config(self):
        config = {
            'param_initializer': initializers.serialize(self.param_initializer),
            'param_regularizer': regularizers.serialize(self.param_regularizer),
            'param_constraint': constraints.serialize(self.param_constraint),
            'shared_axes': self.shared_axes
        }
        base_config = super(ParameterBase, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return self.param.shape


@keras_export('keras.constraints.MinMax')
class MinMax(Constraint):
    """MinMax weight constraint.

    Constrains the weights incident to each hidden unit
    to have values between a lower bound and an upper bound.

    # Arguments
        min_value: the minimum norm for the incoming weights.
        max_value: the maximum norm for the incoming weights.
    """

    def __init__(self, min_value=0.0, max_value=1.0, penalty=1.0):
        self.min_value = min_value
        self.max_value = max_value
        self.penalty = penalty

    def __call__(self, w):
        return MinMaxNorm(self.min_value, self.max_value, self.penalty)(w)

    def get_config(self):
        return {'min_value': self.min_value,
                'max_value': self.max_value}




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from . import math
from . import utilities
from . import validations
from . import activations
from . import initializers
from . import optimizers
from . import callbacks
from . import schedules
# from . import transformers

from .utilities import *
from .validations import *
from .math import *
from .find_neighbors import *
from .callbacks import *


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import tensorflow as tf
tf.get_logger().setLevel('ERROR')
tf.compat.v1.disable_eager_execution()

from . import constraints
from . import functionals
from . import models
from . import utils

from .functionals.functional import Functional
from .functionals.variable import Variable
from .functionals.field import Field
from .functionals.parameter import Parameter
from .models.model import SciModel
from .constraints import Constraint, PDE, Data, Tie

# Also importable from root
from .utils.math import *
from .utils import math
from .utils.utilities import reset_session, clear_session
from .utils.utilities import set_default_log_path, get_default_log_path
from .utils.utilities import set_random_seed
from .utils.utilities import set_floatx
from .utils.utilities import get_bibliography

import json
from urllib import request
from pkg_resources import parse_version

from .references import BIBLIOGRAPHY_STR

# SciANN.
__author__ = "Ehsan Haghighat"
__email__ = "ehsanh@mit.edu"
__copyright__ = "Copyright 2019, Physics-Informed Deep Learning"
__credits__ = []
__url__ = "http://github.com/sciann/sciann]"
__license__ = "MIT"
__version__ = "0.6.7.5"
__cite__ = \
    '@article{haghighat2021sciann, \n' +\
    '    title={SciANN: A Keras/TensorFlow wrapper for scientific computations and physics-informed deep learning using artificial neural networks}, \n' +\
    '    author={Haghighat, Ehsan and Juanes, Ruben}, \n' +\
    '    journal={Computer Methods in Applied Mechanics and Engineering}, \n' +\
    '    year={2021}, \n' +\
    '    url = {https://doi.org/10.1016/j.cma.2020.113552}, \n' +\
    '    howpublished={https://github.com/sciann/sciann.git}, \n' +\
    '}'

# Import message.
_header = '---------------------- {} {} ----------------------'.format(str.upper(__name__), str(__version__))
_footer = len(_header)*'-'
__welcome__ = \
    '{} \n'.format(_header) +\
    'For details, check out our review paper and the documentation at: \n' +\
    ' +  "https://www.sciencedirect.com/science/article/pii/S0045782520307374", \n' +\
    ' +  "https://arxiv.org/abs/2005.08803", \n' +\
    ' +  "https://www.sciann.com". \n' +\
    '\n ' +\
    'Need support or would like to contribute, please join sciann`s slack group: \n' +\
    ' +  "https://join.slack.com/t/sciann/shared_invite/zt-ne1f5jlx-k_dY8RGo3ZreDXwz0f~CeA" ' +\
    '\n \n' +\
    'TensorFlow Version: {} \n'.format(tf.__version__) +\
    'Python Version: {} \n'.format(sys.version)
    # '{} \n'.format(__cite__) +\
    # _footer


import os
if 'SCIANN_WELCOME_MSG' in os.environ.keys() and \
        os.environ['SCIANN_WELCOME_MSG']=='-1':
    pass
else:
    print(__welcome__)

# set default logging directory.
set_default_log_path(os.path.join(os.getcwd(), "logs"))
initialize_bib(BIBLIOGRAPHY_STR)

# check sciann version.
try:
    url = 'https://pypi.python.org/pypi/sciann/json'
    releases = json.loads(request.urlopen(url, timeout=1).read())['releases']
    releases = sorted(releases, key=parse_version, reverse=True)
    if releases.index(__version__) > 0:
        print(f'Outdated SciANN installation is found (V-{__version__}). '
              f'Get the latest version (V-{releases[0]}):  \n '
              f'     > pip [--user] install --upgrade sciann  ')
except:
    pass



""" schedules module define LR schedulers for training
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.keras.callbacks import LearningRateScheduler as LearningRateSchedule
from tensorflow.keras.callbacks import ReduceLROnPlateau
import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


def sci_exponential_decay(initial_learning_rate=1e-3,
                          final_learning_rate=1e-5,
                          decay_epochs=1000,
                          delay_epochs=0,
                          verify=False,
                          scheduler="exponential_decay"):
    """Applies exponential decay to the learning rate.
    Example:
    ``` python
        sci_exponential_decay(1e-3, 1e-5, 10000)
    ```
    Args:
        initial_learning_rate: The initial learning rate.
            Defaults to 1e-3.
        final_learning_rate: The final learning rate.
            Defaults to 1e-5.
        decay_epochs: number of epochs, over which the exponential decay is applied.
            Defaults to 1000.
        delay_epochs: number of epochs, over which the initial learning rate is used.
            Defaults to 0 (No delay).
        verify: Boolean. Plots the learning rate schedule.

    Returns:
        A 1-arg callable learning rate schedule that takes the current optimizer
        step and outputs the decayed learning rate, a scalar `Tensor` of the same
        type as `initial_learning_rate`.
    """
    initial_learning_rate = lr0 = initial_learning_rate
    final_learning_rate = lr1 = final_learning_rate
    decay_epochs = decay_epochs
    decay_rate = np.log(lr1 / lr0) / (decay_epochs)
    epoch_vals = np.linspace(0, decay_epochs, 201)
    lr_vals = lr0 * np.exp(decay_rate * epoch_vals)
    
    # account for delay. 
    if delay_epochs > 0:
        epoch_vals = np.concatenate([[0, delay_epochs], epoch_vals+delay_epochs])
        lr_vals = np.concatenate([[initial_learning_rate, initial_learning_rate], lr_vals])

    if verify is True:
        plot_lr_schedule(epoch_vals, lr_vals)

    func = interp1d(epoch_vals, lr_vals,
                    bounds_error=False,
                    fill_value=(initial_learning_rate, final_learning_rate))

    return LearningRateSchedule(lambda i: float(func(i)))


def sci_sinusoidal_exponential_decay(
        initial_learning_rate=1e-3,
        final_learning_rate=1e-5,
        decay_epochs=1000,
        delay_epochs=0,
        sine_freq=10,
        sine_decay_rate=0.5,
        verify=False,
        scheduler="sinusoidal_exponential_decay"):
    """Applies exponential decay to the learning rate.
    Args:
        initial_learning_rate: The initial learning rate.
            Defaults to 1e-3.
        final_learning_rate: The final learning rate.
            Defaults to 1e-5.
        decay_epochs: number of epochs, over which the exponential decay is applied.
            Defaults to 1000.
        delay_epochs: number of epochs, over which the initial learning rate is used.
            Defaults to 0 (No delay).
        sine_freq: number of sinusoidal oscillations in learning rate.
            defaults to 10.
        sine_decay_rate: exponential decay on the amplitude of sin wave.
            defaults to 0.5.
        verify: Boolean. Plots the learning rate schedule.
    Returns:
        A 1-arg callable learning rate schedule that takes the current optimizer
        step and outputs the decayed learning rate, a scalar `Tensor` of the same
        type as `initial_learning_rate`.
    """
    lr0 = initial_learning_rate
    lr1 = final_learning_rate
    decay_rate = np.log(lr1 / lr0) / (decay_epochs)

    th = np.arctan(decay_rate)

    epoch_vals = np.linspace(0, decay_epochs, 201)
    base_lr_vals = lr0 * np.exp(decay_rate * epoch_vals)
    decay_axis = epoch_vals * np.cos(th) + np.log(base_lr_vals / lr0) * np.sin(th)
    normalized_epoch_vals = 2 * np.pi * epoch_vals / epoch_vals.max()
    sin_lr_vals = np.exp(-sine_decay_rate * normalized_epoch_vals) * \
                  np.sin(-sine_freq * normalized_epoch_vals)
    lr_vals = lr0 * np.exp(decay_axis * np.sin(th) + sin_lr_vals * np.cos(th))

    # account for delay. 
    if delay_epochs > 0:
        epoch_vals = np.concatenate([[0, delay_epochs], epoch_vals+delay_epochs])
        lr_vals = np.concatenate([[initial_learning_rate, initial_learning_rate], lr_vals])

    if verify is True:
        plot_lr_schedule(epoch_vals, lr_vals)

    func = interp1d(epoch_vals, lr_vals,
                    bounds_error=False,
                    fill_value=(initial_learning_rate, final_learning_rate))
    return LearningRateSchedule(lambda i: float(func(i)))


def sci_learning_rate_schedule(
        lr_epochs=None,
        lr_values=None,
        verify=False,
        scheduler="learning_rate_schedule"):
    """uses discrete values to schedule the learning rate.
    Args:
        lr_epochs: list or numpy array of epochs at which learning rate values are given.
        lr_values: List or numpy array of learning rates.
        'sci_learning_rate_schedule'.
        verify: Boolean. Plots the learning rate schedule.
    Returns:
        A 1-arg callable learning rate schedule that takes the current optimizer
        step and outputs the decayed learning rate, a scalar `Tensor` of the same
        type as `initial_learning_rate`.
    """
    func = interp1d(lr_epochs, lr_values,
                    bounds_error=False,
                    fill_value=(lr_values[0], lr_values[-1]))

    if verify is True:
        plot_lr_schedule(lr_epochs, lr_values)

    return LearningRateSchedule(lambda i: float(func(i)))


def setup_lr_scheduler(learning_rate_scheduler):
    assert isinstance(learning_rate_scheduler, dict), \
        'Expecting a dictionary describing the scheduler. '
    if 'scheduler' in learning_rate_scheduler.keys():
        if learning_rate_scheduler['scheduler'].lower() in \
                alternative_names(["exponential", "decay"]):
            try:
                return sci_exponential_decay(**learning_rate_scheduler)
            except:
                help(sci_exponential_decay)
                raise ValueError('Inconsistent inputs for `sci_exponential_decay`.')
        elif learning_rate_scheduler['scheduler'].lower() in \
                alternative_names(["sine", "exponential", "decay"]):
            try:
                return sci_sinusoidal_exponential_decay(**learning_rate_scheduler)
            except:
                help(sci_sinusoidal_exponential_decay)
                raise ValueError('Inconsistent inputs for `sci_sinusoidal_exponential_decay`.')
        elif learning_rate_scheduler['scheduler'].lower() in \
                alternative_names(["learning", "rate", "scheduler"]):
            try:
                return sci_learning_rate_schedule(**learning_rate_scheduler)
            except:
                help(sci_learning_rate_schedule)
                raise ValueError('Inconsistent inputs for `sci_learning_rate_schedule`.')
        elif learning_rate_scheduler['scheduler'].lower() in ('default',):
            try:
                return ReduceLROnPlateau(
                    monitor='loss', factor=0.5,
                    patience=learning_rate_scheduler['reduce_lr_after'],
                    verbose=1, mode='auto',
                    min_delta=learning_rate_scheduler['reduce_lr_min_delta'],
                    min_lr=0.
                )
            except:
                raise ValueError(
                    'Incorrect values for `reduce_lr_after` or `reduce_lr_min_delta`. '
                )

        else:
            raise ValueError('Unrecognized scheduler.')
    else:
        raise ValueError('`scheduler` not found in the dictionary. ')


def alternative_names(name):
    assert isinstance(name, list)
    assert all([isinstance(v, str) for v in name])
    alts = [
        "".join(name).lower(),
        "_".join(name).lower(),
        "".join([v[0] for v in name]).lower(),
    ]
    return alts


def plot_lr_schedule(epochs, lrs):
    """ Plotting (epochs, lr) for debugging purposes.
    Args:
        epochs: list or numpy array of epochs at which learning rate values are given.
        lrs: List or numpy array of learning rates.
    """
    fig, ax = plt.subplots(1, 2, figsize=(10, 4))
    ax[0].plot(epochs, lrs)
    ax[0].set_xlabel("epochs")
    ax[0].set_ylabel("LR Schedule")

    ax[1].semilogy(epochs, lrs)
    ax[1].set_xlabel("epochs")
    ax[1].set_ylabel("LR Schedule")
    plt.subplots_adjust(0.15, 0.15, 0.85, 0.85, 0.3, 0.2)

    plt.show()
    plt.close(fig)



""" Process/Prepare standard data for the `Radial Basis Networks`.
"""

import numpy as np
from sklearn.neighbors import BallTree


def find_neighbors(*args, **kwargs):
    """ Find closest point to each node in xs.

    # Arguments
        xs: A list of input vectors of dimension (N,1) with N as number of data-points.
        ys: A list of output vectors of dimension (N,1) with N as number of data-points.
        size_rb: Number of closest points.
        return_ids: Boolean.

    # Returns
        xrb: A list of N-`neighbors` of dimension (N,size_rb).
    """
    if len(args)==2:
        xs = args[0]
        size_rb = args[1]
    elif len(args)==3:
        xs = [args[0], args[1]]
        size_rb = args[2]
    else:
        raise ValueError

    # prepare data.
    if all([x.shape[-1]!=1 for x in xs]):
        for i, x in enumerate(xs):
            xs[i] = x.reshape(x.shape + (1,))
    xsc = np.concatenate(xs, axis=-1)

    tree = BallTree(xsc, size_rb)
    ids = tree.query(xsc, size_rb, return_distance=False)

    if 'return_ids' in kwargs and kwargs['return_ids']:
        return ids
    else:
        return [x[ids].reshape(-1, size_rb) for x in xs]




'''
# Curve fitting in 1D

Here, a 1D curve fitting example is explored. Imagine, a synthetic data
generated from \\\( \sin(x) \\\) over the range of \\\( [0, 2\pi] \\\).

To train a neural network model on this curve, you should first define a `Variable`.

A neural network with three layers, each containing 10 neurons, and with `tanh` activation function is then generated
using the `Functional` class.

The target is imposed on the output using the `Data` class from `Constraint`, and passed to the `SciModel` to form a
SciANN model.
'''

import numpy as np
from sciann import Variable, Functional, SciModel, Parameter
from sciann.constraints import Data, MinMax
import sciann as sn 
import time


# Synthetic data generated from sin function over [0, 2pi]
x_true = np.linspace(0, np.pi*2, 10000)
y_true = np.sin(x_true)
dy_true = np.cos(x_true)

# The network inputs should be defined with Variable.
x = Variable('x')
xf = Functional('xf', x)
xf.set_trainable(False)

# Each network is defined by Functional.
y = Functional('y', x, [10, 10, 10], activation=['tanh', 'g-cos', 'l-sin'])
dy_dx = sn.diff(y, x)

d = Parameter(2.0, inputs=x)

# Define the target (output) of your model.
c1 = Data(y)

# The model is formed with input `x` and condition `c1`.
model = SciModel(x, [y, dy_dx], optimizer='adam')
model.summary()

start_time = time.time()

# Training: .train runs the optimization and finds the parameters.
model.train(x_true,
            [y_true, dy_true],
            epochs=100,
            learning_rate={"scheduler": "ExponentialDecay",
                           "initial_learning_rate": 1e-3,
                           "final_learning_rate": 1e-5,
                           "decay_epochs": 10,
                           "verify": False},
            batch_size=32,
            adaptive_weights={'method': "SA", "eta": 0.01}
            )

print(f"Training finished in {time.time()-start_time}s. ")

# used to evaluate the model after the training.
y_pred = y.eval(model, x_true)

# print(x_true.shape, y_pred.shape)
import matplotlib.pyplot as plt
plt.plot(x_true, y_true, x_true, y_pred)
plt.show()



from tensorflow import math as tf_math
from tensorflow.python.util.tf_export import keras_export
import keras as k
from keras import constraints
from keras import initializers
from keras import regularizers
from keras.utils import tf_utils
from tensorflow.python.keras.layers import InputSpec
from keras.initializers import Constant as default_constant_initializer


@keras_export('keras.activations.SciActivation')
class SciActivation(k.layers.Activation):
    """Applies an activation function to an output.

    # Arguments
        w0: The factor to be applied to initialized weights.
            e.g. sin(w0 * input)
        activation: name of activation function to use
            (see: [activations](../activations.md)),
            or alternatively, a TensorFlow operation.

    # Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    # Output shape
        Same shape as input.
    """

    def __init__(self, w0=1.0, activation='linear', **kwargs):
        super(SciActivation, self).__init__(
            get_activation(activation),
            **kwargs
        )
        self.activation_name = self.activation.__name__
        self.w0 = w0

    def call(self, inputs):
        return self.activation(self.w0 * inputs)

    def get_config(self):
        config = {'w0': self.w0}
        base_config = super(SciActivation, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


@keras_export('keras.layers.SciActivationLayer')
class SciActivationLayer(k.layers.Layer):
    """Applies an activation layer with trainable parameter to an output.

    # Arguments
        alpha: The trainable factor to be applied to initialized weights.
            e.g. sin(alpha * input)
        activation: name of activation function to use
            (see: [activations](../activations.md)),
            or alternatively, a TensorFlow operation.
        type: ('l', 'g') for locally or globally activation functions.

    # Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    # Output shape
        Same shape as input.
    """

    def __init__(self, alpha=1.0, activation='linear', type='l', **kwargs):
        super(SciActivationLayer, self).__init__(**kwargs)
        self.activation = get_activation(activation)
        self.activation_name = self.activation.__name__
        self.alpha_initializer = default_constant_initializer(alpha)
        self.alpha_regularizer = None
        self.alpha_constraint = None
        self.shared_axes = None
        assert type in ('l', 'g')
        self.type = type

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        param_shape = list(input_shape[1:])
        if self.shared_axes is not None:
            for i in self.shared_axes:
                param_shape[i - 1] = 1
        self.alpha = self.add_weight(
            shape=param_shape if self.type=='l' else [1,],
            name='alpha',
            initializer=self.alpha_initializer,
            regularizer=self.alpha_regularizer,
            constraint=self.alpha_constraint)
        # Set input spec
        axes = {}
        if self.shared_axes:
            for i in range(1, len(input_shape)):
                if i not in self.shared_axes:
                    axes[i] = input_shape[i]
        self.built = True

    def call(self, inputs):
        return self.activation(self.alpha * inputs)

    def get_config(self):
        config = {
            'alpha_initializer': initializers.serialize(self.alpha_initializer),
            'alpha_regularizer': regularizers.serialize(self.alpha_regularizer),
            'alpha_constraint': constraints.serialize(self.alpha_constraint),
            'shared_axes': self.shared_axes
        }
        base_config = super(SciActivationLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape


@keras_export('keras.layers.SciRowdyActivationLayer')
class SciRowdyActivationLayer(k.layers.Layer):
    """Adds multiple activations together.

    # Arguments
        activation: list of activations functions of class SciActivation or SciActivationLayer.

    # Output shape
        Same shape as input.
    """

    def __init__(self, phis, **kwargs):
        super(SciRowdyActivationLayer, self).__init__(**kwargs)
        self.phis = phis
        self.activation = None
        self.activation_name = "rowdy"
        self.alpha_initializer = default_constant_initializer(1.0/len(phis))
        self.alpha_regularizer = None
        self.alpha_constraint = None

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        for i, phi in enumerate(self.phis):
            self.add_weight(
                shape=[1, ],
                name=f'alpha{i}',
                initializer=self.alpha_initializer,
                regularizer=self.alpha_regularizer,
                constraint=self.alpha_constraint)
        self.built = True

    def call(self, inputs):
        sums = [a*f(inputs) for a, f in zip(self.weights, self.phis)]
        return tf_math.add_n(sums, name='rowdy_sum')

    def get_config(self):
        config = {
            'alpha_initializer': initializers.serialize(self.alpha_initializer),
            'alpha_regularizer': regularizers.serialize(self.alpha_regularizer),
            'alpha_constraint': constraints.serialize(self.alpha_constraint),
            'shared_axes': self.shared_axes
        }
        base_config = super(SciRowdyActivationLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape


def get_activation(activation):
    """ Evaluates the activation function from a string or list of string inputs.

    # Arguments
        activation: A string pointing to the function name.

    # Returns:
        A function handle.
    """

    if isinstance(activation, list):
        return [get_activation(act) for act in activation]

    elif isinstance(activation, str):
        if hasattr(k.activations, activation):
            return getattr(k.activations, activation)
        elif hasattr(k.backend, activation):
            return getattr(k.backend, activation)
        elif hasattr(tf_math, activation):
            return getattr(tf_math, activation)
        else:
            raise ValueError(
                'Not a valid function name: ' + activation +
                ' - Please provide a valid activation '  
                'function name from tensorflow.python.keras or Tensorflow. '
            )

    elif callable(activation):
        return activation

    else:
        raise TypeError(
            'Please provide a valid input: ' + type(activation) +
            ' - Expecting a function name or function handle. '
        )


k.utils.generic_utils.get_custom_objects().update({
    'SciActivation': SciActivation
})



""" Data class to impose data constraint.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .constraint import Constraint
from ..utils import is_functional


class Data(Constraint):
    """ Data class to impose to the system.

    # Arguments
        cond: Functional.
            The `Functional` object that Data condition
            will be imposed on.
        name: String.
            A `str` for name of the pde.

    # Returns

    # Raises
        ValueError: 'cond' should be a functional object.
                    'mesh' should be a list of numpy arrays.
    """
    def __init__(self, cond, name="data", **kwargs):
        if not is_functional(cond):
            raise ValueError(
                "Expected a Functional object as the `cond`, received a "
                "{} - {}".format(type(cond), cond)
            )
        if 'x_true_ids' in kwargs or 'y_true' in kwargs:
            raise ValueError(
                "Legacy inputs: please check `SciModel` and `SciModel.train` on how to impose partial data. "
            )
        elif len(kwargs)>0:
            raise ValueError(
                'Unrecognized input variable: {}'.format(kwargs.keys())
            )
        super(Data, self).__init__(
            cond=cond,
            name=name
        )



""" PDE class to impose pde constraint.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .constraint import Constraint
from ..utils import is_functional


class PDE(Constraint):
    """ PDE class to impose to the system.

    # Arguments
        pde: Functional.
            The `Functional` object that pde if formed on.
        name: String.
            A `str` for name of the pde.

    # Returns

    # Raises
        ValueError: 'pde' should be a functional object.
    """
    def __init__(self, pde, name="pde", **kwargs):
        if not is_functional(pde):
            raise ValueError(
                "Expected a Functional object as the pde, received a "
                "{} - {}".format(type(pde), pde)
            )
        # Check inputs.
        if 'mesh_ids' in kwargs or 'sol' in kwargs:
            raise ValueError(
                'Legacy interface: please use `SciModel` and `SciModel.train` to impose on specific ids. '
            )
        elif len(kwargs)>0:
            raise ValueError(
                'Unrecognized input variable: {}'.format(kwargs.keys())
            )
        super(PDE, self).__init__(
            cond=pde,
            name=name
        )



""" Functional class for SciANN.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras import backend as K
graph_unique_name = K.get_graph().unique_name

from keras.layers import Dense
from keras.layers import Activation
from keras.layers import Concatenate
from keras.layers import Lambda
# from keras.layers import BatchNormalization
from keras.models import Model
from tensorflow import tensordot, expand_dims

from ..utils import to_list, unpack_singleton, is_same_tensor, unique_tensors
from ..utils import default_weight_initializer
from ..utils import default_regularizer
from ..utils import validations, getitem
from ..utils import floatx, set_floatx
from ..utils import math
from ..utils.activations import SciActivation, get_activation
from ..utils import prepare_default_activations_and_initializers

from .field import Field
from .mlp_functional import MLPFunctional


""" Configures the Functional object (Neural Network).

# Arguments
    fields: String or Field.
        [Sub-]Network outputs.
        It can be of type `String` - Associated fields will be created internally.
        It can be of type `Field` or `Functional`
    variables: Variable.
        [Sub-]Network inputs.
        It can be of type `Variable` or other Functional objects.
    hidden_layers: A list indicating neurons in the hidden layers.
        e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
    activation: defaulted to "tanh".
        Activation function for the hidden layers.
        Last layer will have a linear output.
    output_activation: defaulted to "linear".
        Activation function to be applied to the network output.
    res_net: (True, False). Constructs a resnet architecture.
        Defaulted to False.
    kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
    bias_initializer: Initializer of the `Bias`, from `k.initializers`.
    kernel_regularizer: Regularizer for the kernel.
        To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
    bias_regularizer: Regularizer for the bias.
        To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
    trainable: Boolean.
        False if network is not trainable, True otherwise.
        Default value is True.

# Raises
    ValueError:
    TypeError:
"""
def Functional(
        fields=None,
        variables=None,
        hidden_layers=None,
        activation="tanh",
        output_activation="linear",
        res_net=False,
        kernel_initializer=None,
        bias_initializer=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        trainable=True,
        **kwargs):
    # prepare hidden layers.
    if hidden_layers is None:
        hidden_layers = []
    else:
        hidden_layers = to_list(hidden_layers)
    if not all([isinstance(n, int) for n in hidden_layers]):
        raise TypeError("Enter a list of integers as the third input assigning layer widths, e.g. [10,10,10]. ")
    # prepare kernel initializers.
    activations, def_biasinit, def_kerinit = \
        prepare_default_activations_and_initializers(
        len(hidden_layers) * [activation] + [output_activation]
    )
    if kernel_initializer is None:
        kernel_initializer = def_kerinit
    elif isinstance(kernel_initializer, (float, int)):
        kernel_initializer = default_weight_initializer(
            len(hidden_layers) * [activation] + [output_activation],
            'constant',
            scale=kernel_initializer
        )
    else:
        kernel_initializer = [kernel_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
    # prepare bias initializers.
    if bias_initializer is None:
        bias_initializer = def_biasinit
    elif isinstance(bias_initializer, (float, int)):
        bias_initializer = default_weight_initializer(
            len(hidden_layers) * [activation] + [output_activation],
            'constant',
            scale=bias_initializer
        )
    else:
        bias_initializer = [bias_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
    # prepare regularizers.
    kernel_regularizer = default_regularizer(kernel_regularizer)
    bias_regularizer = default_regularizer(bias_regularizer)
    # prepares fields.
    output_fields = []
    for fld in to_list(fields):
        if isinstance(fld, (str, tuple, list)):
            if isinstance(fld, str):
                fld_name, fld_units = Field.prepare_field_inputs(fld)
            else:
                fld_name, fld_units = Field.prepare_field_inputs(*fld)
            output_fields.append(
                Field(
                    name=fld_name,
                    units=fld_units,
                    kernel_initializer=kernel_initializer[-1],
                    bias_initializer=bias_initializer[-1],
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                )
            )
        elif validations.is_field(fld):
            output_fields.append(fld)
        else:
            raise TypeError(
                'Please provide a "list" of field names of'
                + ' type "String" or "Field" objects.'
            )
    # prepare inputs/outputs/layers.
    inputs = []
    layers = []
    variables = to_list(variables)
    if all([isinstance(var, MLPFunctional) for var in variables]):
        for var in variables:
            inputs += var.outputs
        # for var in variables:
        #     for lay in var.layers:
        #         layers.append(lay)
    else:
        raise TypeError(
            "Input error: Please provide a `list` of `sn.Variable`s. \n"
            "Provided - {}".format(variables)
        )

    # Input layers.
    if len(inputs) == 1:
        net_input = inputs[0]
    else:
        layer = Concatenate(name=graph_unique_name('concat'))
        net_input = layer(inputs)

    # Define the output network.
    net = [net_input]

    # define the ResNet networks.
    if res_net is True:
        res_layers = []
        res_outputs = []
        for rl in ["U", "V", "H"]:
            layers.append(
                Dense(
                    hidden_layers[0],
                    kernel_initializer=kernel_initializer[0],
                    bias_initializer=bias_initializer[0],
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    name=graph_unique_name("DRes"+rl+"{:d}b".format(hidden_layers[0])),
                    **kwargs
                )
            )
            res_output = layers[-1](net_input)
            # Apply the activation.
            if activations[0].activation_name != 'linear':
                layers.append(activations[0])
                res_outputs.append(layers[-1](res_output))
        net[-1] = res_outputs[-1]

    for nLay, nNeuron in enumerate(hidden_layers):
        # Add the layer.
        layer = Dense(
            nNeuron,
            kernel_initializer=kernel_initializer[nLay],
            bias_initializer=bias_initializer[nLay],
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            trainable=trainable,
            name=graph_unique_name("D{:d}b".format(nNeuron)),
            **kwargs
        )
        layers.append(layer)
        net[-1] = layer(net[-1])
        # # Add batch normalization. 
        # layer = BatchNormalization()
        # layers.append(layer)
        # net[-1] = layer(net[-1])
        # Apply the activation.
        if activations[nLay].activation_name != 'linear':
            layer = activations[nLay]
            layers.append(layer)
            net[-1] = layer(net[-1])
        # Add the resnet layer
        if res_net is True:
            layer = Lambda(lambda xs: (1-xs[0])*xs[1] + xs[0]*xs[2], name=graph_unique_name("ResLayer"))
            net[-1] = layer([net[-1]] + res_outputs[:2])

    # Assign to the output variable
    if len(net) == 1:
        net_output = net[0]
    else:
        raise ValueError("Legacy for Enrichment: Must be updated. ")
        layer = Concatenate(name=graph_unique_name('concat'))
        net_output = layer(net)

    # Define the final outputs of each network
    functionals = []
    for out in output_fields:
        last_layers = [out]
        last_output = out(net_output)
        # add the activation on the output.
        if activations[-1].activation_name != 'linear':
            layer = activations[-1]
            last_layers.append(layer)
            last_output = layer(last_output)
        # Construct functionals
        functionals.append(
            MLPFunctional(inputs, [last_output], layers + last_layers)
        )

    return unpack_singleton(functionals)




from setuptools import setup
from setuptools import find_packages

long_description = '''
SciANN is an Artificial Neural Netowork library, 
based on Python, Keras, and TensorFlow, designed
to perform scientific computations, solving ODEs 
and PDEs, curve-fitting, etc, very efficiently.

Read the documentation at: https://sciann.com

SciANN is compatible with Python 2.7-3.6
and is distributed under the MIT license.
'''

setup(
    name='SciANN',
    version='0.6.7.5',
    description='A Keras/Tensorflow wrapper for scientific computations and physics-informed deep learning using artificial neural networks.',
    long_description=long_description,
    author='Ehsan Haghighat',
    author_email='ehsan@sciann.com',
    license='MIT',
    url='https://github.com/sciann/sciann',
    install_requires=['numpy',
                      'scipy',
                      'six',
                      'pyyaml',
                      'h5py',
                      'sklearn',
                      'pybtex',
                      'tensorflow>=2.6.0,<=2.9.5',
                      ],
    extras_require={
          'visualize': ['pydot>=1.2.4', 'matplotlib'],
          'tests': ['pytest',
                    'pytest-pep8',
                    'pytest-xdist',
                    'flaky',
                    'pytest-cov',
                    'pandas',
                    'requests',
                    'graphviz',
                    'pydot',
                    'markdown',
                    'matplotlib',
                    'pyux',
                    ],
    },
    classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
    packages=find_packages(),
    include_package_data=True
)



""" Utilities to process functionals.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf 
from keras import backend as K
graph_unique_name = K.get_graph().unique_name

from keras.layers import Layer, Lambda, Flatten, Dense
from keras.layers import Dot
from keras.layers import Input
from keras.models import Model
from tensorflow import sin as tf_sin
from tensorflow import cos as tf_cos
from tensorflow import tile as tf_tile
from tensorflow import pow as tf_pow
from tensorflow import divide as tf_divide

from tensorflow import gradients as tf_gradients
from tensorflow import stop_gradient as tf_stop_gradient
from tensorflow import multiply, expand_dims

from .utilities import *
from .validations import *


def fourier(f, w=10):
    """Apply Fourier transform to the `Variable` objects..

    # Arguments
        w: (Int, np.ndarray) Frequencies of transformation.

    # Returns
        A Functional.
    """
    validate_functional(f)
    append_to_bib("wang2020eigenvector")
    layers = []
    outputs = []
    for fi in f.outputs:
        if isinstance(w, int):
            w = 2*np.pi/np.random.rand(w)
        else:
            assert isinstance(w, np.ndarray)
        size_i = fi.shape.as_list()
        assert len(size_i) == 2, 'Designed for the MLP model.'
        assert size_i[-1] == 1, 'Only supported now.'
        layers.append(
            Dense(
                w.size,
                use_bias=False,
                trainable=False,
                activation=tf_sin,
                name=graph_unique_name("fourier-sin")
            )
        )
        outputs.append(layers[-1](fi))
        layers[-1].set_weights(
            [w.reshape(layers[-1].weights[0].shape.as_list())]
        )
        layers.append(
            Dense(
                w.size,
                use_bias=False,
                trainable=False,
                activation=tf_cos,
                name=graph_unique_name("fourier-cos")
            )
        )
        outputs.append(layers[-1](fi))
        layers[-1].set_weights(
            [w.reshape(layers[-1].weights[0].shape.as_list())]
        )

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(f.inputs.copy()),
        outputs=outputs,
        layers=layers
    )

    return res


def monomial(f, p=10, w=1.):
    """Apply monomial feature transformation to the `Variable` objects..

    # Arguments
        p: (Int, list) monomial powers to be considered.
        w: (Int, list) weights for each term in the monomial(e.g. 1/n! for taylor expansion).

    # Returns
        A Functional.
    """
    validate_functional(f)
    if isinstance(p, int):
        p = list(range(1, p+1))
    else:
        assert isinstance(p, (np.ndarray, list))
    if isinstance(w, float):
        w = len(p)*[w]
    else:
        assert isinstance(w, (np.ndarray, list))
    layers = []
    outputs = []
    for fi in f.outputs:
        f_dim = fi.shape.as_list()
        tile_dim = (len(f_dim)-1)*[1] + [len(p)]
        layers.append(
            Lambda(lambda ys: tf_divide(tf_pow(tf_tile(ys, tile_dim), p), w),
                   name=graph_unique_name("monomials"))
        )
        outputs.append(
            layers[-1](fi)
        )

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(f.inputs.copy()),
        outputs=outputs,
        layers=layers
    )

    return res


def outer(a, b):
    """outer product of two `Functional` objects.

    # Arguments
        a, b: outer(a,b)
        Note that at least one of them should be of type Functional.

    # Returns
        A Functional.
    """
    validate_functional(a)
    validate_functional(b)
    layers = []
    outputs = []
    for a_out in a.outputs:
        for b_out in b.outputs:
            a_shape = a_out.shape.as_list()
            b_shape = b_out.shape.as_list()
            a_exp = len(a_shape)
            b_exp = 1 if b_shape[0] is None else 0
            name = graph_unique_name("outer")
            layers.append(
                Lambda(lambda ys: multiply(expand_dims(ys[0], a_exp),
                                           expand_dims(ys[1], b_exp)), name=name)
            )
            net_output = layers[-1]([a_out, b_out])
            layers.append(Flatten())
            outputs.append(layers[-1](net_output))
    # return the functional
    assert a.get_class() == b.get_class()
    Functional = a.get_class()
    res = Functional(
        inputs=unique_tensors(a.inputs.copy() + b.inputs.copy()),
        outputs=outputs,
        layers=layers
    )
    return res


def pow(a, b):
    """Element-wise exponentiation applied to the `Functional` object.

    # Arguments
        a, b: pow(a,b)
        Note that at least one of them should be of type Functional.

    # Returns
        A Functional.
    """
    if is_functional(a):
        validate_functional(a)
        f, p = a, b
        name = "pow{:d}".format(p) if isinstance(p, int) else "pow{:.3f}".format(p)
        lmbd = [Lambda(lambda x: x ** p, name=graph_unique_name(name)) for X in f.outputs]

    elif is_functional(b):
        validate_functional(b)
        f, p = b, a
        name = "pow{:d}".format(p) if isinstance(p, int) else "pow{:.3f}".format(p)
        lmbd = [Lambda(lambda x: p ** x, name=graph_unique_name(name)) for X in f.outputs]

    else:
        raise ValueError('Expected one functional in the arguments.')

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(f.inputs.copy()),
        outputs = _apply_operation(lmbd, f),
        layers = lmbd
    )

    return res


def add(f, other):
    """Element-wise addition applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[0]+x[1], name=graph_unique_name("add")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: x+other, name=graph_unique_name("add")) for X in f.outputs]
    
    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def radd(f, other):
    """Element-wise right-addition applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    return add(f, other)


def sub(f, other):
    """Element-wise subtraction applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[0]-x[1], name=graph_unique_name("sub")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: x-other, name=graph_unique_name("sub")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def rsub(f, other):
    """Element-wise right-subtraction applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[1]-x[0], name=graph_unique_name("rsub")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: other-x, name=graph_unique_name("rsub")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def mul(f, other):
    """Element-wise multiplication applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[0]*x[1], name=graph_unique_name("mul")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: x*other, name=graph_unique_name("mul")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def rmul(f, other):
    """Element-wise right-multiplication applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    return mul(f, other)


def div(f, other):
    """Element-wise division applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[0]/x[1], name=graph_unique_name("div")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: x/other, name=graph_unique_name("div")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def rdiv(f, other):
    """Element-wise right-division applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: x[1]/x[0], name=graph_unique_name("rdiv")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: other/x, name=graph_unique_name("rdiv")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = _apply_operation(lmbd, f, other),
        layers = lmbd
    )
    return res


def dot(f, other):
    """Dot product of two `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)
    validate_functional(other)
    assert len(f.outputs) == len(other.outputs)

    outputs = []
    layers = []
    for fl, fr in zip(f.outputs, other.outputs):
        assert fl.shape.as_list() == fr.shape.as_list(),\
            'Expected equal dimensions for output of functionals. '
        l = Lambda(
            lambda x: K.reshape(tf.math.reduce_sum(x*fr, list(range(1, len(fl.shape)))), [-1, 1]),
            name=graph_unique_name("dot")
        )
        layers += [l]
        outputs += [l(fl)]
        
    inputs = to_list(f.inputs) + to_list(other.inputs)
    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = outputs,
        layers = layers
    )
    return res


def diag_part(f):
    """Diag_part operation returns diagonal part of outputs of (None,N,N) functional.

    # Arguments
        f: Functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    lmbd = []
    outputs = []
    for o in f.outputs:
        assert len(o.shape) == 3, \
            'Exptected output dimension to be (None, N, N)'
        dim = o.shape[-1]
        l = Lambda(
            lambda x: tf.linalg.diag_part(x),
            name=graph_unique_name("diag_part")
        )
        lmbd += [l]
        outputs += [l(o)]

    Functional = f.get_class()
    res = Functional(
        inputs = f.inputs.copy(),
        outputs = outputs,
        layers = lmbd
    )
    return res


def diag(f):
    """Diag operation converts a vector output (None, N) to a matrix form of (None,N,N) functional.

    # Arguments
        f: Functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    lmbd = []
    outputs = []
    for o in f.outputs:
        assert len(o.shape) == 2, \
            'Exptected output dimension to be (None, N)'
        dim = o.shape[-1]
        l = Lambda(
            lambda x: tf.linalg.diag(x),
            name=graph_unique_name("diag")
        )
        lmbd += [l]
        outputs += [l(o)]

    Functional = f.get_class()
    res = Functional(
        inputs = f.inputs.copy(),
        outputs = outputs,
        layers = lmbd
    )
    return res


def _apply_operation(lambda_layer, lhs, rhs=None):
    """Element-wise mathematical operation applied on the `Functional` objects.

    # Arguments
        lambda_layer: the layers to perform the operation.
        lhs: left hand side objects.
        rhs: right hand side objects.

    # Returns
        output tensors.
    """
    validate_functional(lhs)

    if is_functional(rhs):
        outputs = [l([x, y]) for l, x, y in zip(lambda_layer, lhs.outputs, rhs.outputs)]
    else:
        try:
            outputs = [l(x) for l, x in zip(lambda_layer, lhs.outputs)]
        except (ValueError, TypeError):
            print(
                'Unsupported operation with an object of type {}. '.format(type(lhs))
            )
            outputs = None

    return outputs


def sin(x):
    """Computes sin of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'sin')


def asin(x):
    """Computes asin of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'asin')


def cos(x):
    """Computes cos of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'cos')


def acos(x):
    """Computes acos of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'acos')


def tan(x):
    """Computes tan of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'tan')


def atan(x):
    """Computes atan of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'atan')


def cot(x):
    """Computes cot of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'cot')


def acot(x):
    """Computes acot of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'acot')


def sinh(x):
    """Computes sinh of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'sinh')


def cosh(x):
    """Computes cosh of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'cosh')


def tanh(x):
    """Computes tanh of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'tanh')


def coth(x):
    """Computes coth of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'coth')


def sigmoid(x):
    """Computes coth of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'sigmoid')


def abs(x):
    """Computes abs of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'abs')


def sign(x):
    """Computes sign of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'sign')


def step(x):
    """Computes step (Heaviside) of x element-wise.
       H(x) = 0 if x<=0
       H(x) = 1 if x>0

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    validate_functional(x)

    lmbd = []
    for i in range(len(x.outputs)):
        lmbd.append(
            Lambda(
                lambda x: K.cast(K.greater(x, 0.0), x.dtype), 
                name=graph_unique_name("step")
            )
        )
        
    Functional = x.get_class()
    res = Functional(
        inputs = x.inputs.copy(),
        outputs = _apply_operation(lmbd, x),
        layers = lmbd
    )
    return res


def log(x):
    """Computes log of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'log')


def log10(x):
    """Computes log10 of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return log(x)/np.log(10.0)


def exp(x):
    """Computes exp of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'exp')


def sqrt(x):
    """Computes sqrt of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'sqrt')


def square(x):
    """Computes square of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'square')


def relu(x):
    """Computes relu of x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    return _apply_function(x, 'relu')


def mean(x, **kwargs):
    """Apply mean to the `Functional` objects on far-right axis.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    if "axis" not in kwargs:
        kwargs["axis"] = -1
    if "keepdims" not in kwargs:
        kwargs["keepdims"] = True
    return _apply_function(x, 'mean', **kwargs)


def equal(f, other, tol=None):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.
        tol: (float) If you need a tolerance measure.

    # Returns
        A Functional.
    """
    validate_functional(f)
    assert isinstance(tol, (type(None), float)), 'Expected a floating value for `tol`.'

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        if tol is None:
            lambda_opr = lambda x: K.cast_to_floatx(K.equal(x[0], x[1]))
        else:
            lambda_opr = lambda x: K.cast_to_floatx(K.less_equal(K.abs(x[0]-x[1]), tol))
    else:
        _warn_for_ndarray(other)
        if tol is None:
            lambda_opr = lambda x: K.cast_to_floatx(K.equal(x, other))
        else:
            lambda_opr = lambda x: K.cast_to_floatx(K.less_equal(K.abs(x-other), tol))

    lmbd = [Lambda(lambda_opr, name=graph_unique_name("equal")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


def not_equal(f, other, tol=None):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.
        tol: (float) If you need a tolerance measure.

    # Returns
        A Functional.
    """
    validate_functional(f)
    assert isinstance(tol, (type(None), float)), 'Expected a floating value for `tol`.'

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        if tol is None:
            lambda_opr = lambda x: K.cast_to_floatx(K.not_equal(x[0], x[1]))
        else:
            lambda_opr = lambda x: K.cast_to_floatx(K.greater(K.abs(x[0] - x[1]), tol))
    else:
        _warn_for_ndarray(other)
        if tol is None:
            lambda_opr = lambda x: K.cast_to_floatx(K.not_equal(x, other))
        else:
            lambda_opr = lambda x: K.cast_to_floatx(K.greater(K.abs(x - other), tol))

    lmbd = [Lambda(lambda_opr, name=graph_unique_name("not_equal")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


def greater(f, other):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater(x[0], x[1])), name=graph_unique_name("greater")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater(x, other)), name=graph_unique_name("greater")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


# legacy support
tol_equal = equal


def greater_equal(f, other):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater_equal(x[0], x[1])), name=graph_unique_name("greater_equal")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.greater_equal(x, other)), name=graph_unique_name("greater_equal")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


def less(f, other):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less(x[0], x[1])), name=graph_unique_name("less")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less(x, other)), name=graph_unique_name("less")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


def less_equal(f, other):
    """Element-wise comparison applied to the `Functional` objects.

    # Arguments
        f: Functional object.
        other: A python number or a tensor or a functional object.

    # Returns
        A Functional.
    """
    validate_functional(f)

    inputs = f.inputs.copy()
    if is_functional(other):
        inputs += to_list(other.inputs)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less_equal(x[0], x[1])), name=graph_unique_name("less_equal")) for X in f.outputs]
    else:
        _warn_for_ndarray(other)
        lmbd = [Lambda(lambda x: K.cast_to_floatx(K.less_equal(x, other)), name=graph_unique_name("less_equal")) for X in f.outputs]

    Functional = f.get_class()
    res = Functional(
        inputs=unique_tensors(inputs),
        outputs=_apply_operation(lmbd, f, other),
        layers=lmbd
    )
    return res


def _apply_function(x, fname, **kwargs):
    """Apply `fname` function to x element-wise.

    # Arguments
        x: Functional object.

    # Returns
        A new functional object.
    """
    validate_functional(x)

    fun = get_activation(fname)
    lmbd = []
    for i in range(len(x.outputs)):
        lmbd.append(
            Lambda(
                lambda x: fun(x, **kwargs),
                name=graph_unique_name("{}".format(fname))
            )
        )
    Functional = x.get_class()
    res = Functional(
        inputs = x.inputs.copy(),
        outputs = _apply_operation(lmbd, x),
        layers = lmbd
    )
    return res


def getitem(x, item):
    """returns specific item of a tensor (Functional).

    # Arguments
        item: Item list.

    # Returns
        A new functional object.
    """
    validate_functional(x)

    in_item = item
    print(in_item)
    if not isinstance(in_item, tuple):
        in_item = (in_item,)
    print(in_item)

    itms = (slice(None, None, None),)
    for it in in_item:
        itms += (slice(it, it+1) if isinstance(it, int) else it, )
    
    lmbd = []
    ys = []
    for y in x.outputs:
        l = Lambda(
            lambda xx: xx[itms], 
            name=graph_unique_name("slice")
        )
        lmbd.append(l)
        ys.append(l(y))

    Functional = x.get_class()
    res = Functional(
        inputs = x.inputs.copy(),
        outputs = ys,
        layers = lmbd
    )
    return res


def _gradients(ys, xs, order=1):
    """Returns the gradients of y in `ys` w.r.t. x in `xs`.

    `ys` and `xs` are each a Tensor or a list of tensors.

    # Arguments
        ys: A tensor or list of tesnors to be differentiated.
        xs: A tensor or list of tensors to be used for differentiation.
        order: Order of differentiation.

    # Returns
        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
    """
    if ys.shape[-1] == 1:
        ds = ys
        for i in range(order):
            ds = unpack_singleton(
                tf_gradients(
                    ds, xs,
                    unconnected_gradients='zero',
                    # colocate_gradients_with_ops=True, TF: V1.14.0
                )
            )

    else:
        splitted_ys = tf.split(ys, num_or_size_splits=ys.shape[-1], axis=-1)
        ds = []
        for j, y in enumerate(splitted_ys):
            ds.append(y)
            for i in range(order):
                ds[-1] = unpack_singleton(
                    tf_gradients(
                        ds[-1], xs,
                        unconnected_gradients='zero',
                        # colocate_gradients_with_ops=True, TF: V1.14.0
                    )
                )
            new_shape = [x if x is not None else -1 for x in ds[-1].shape + (1,)]
            ds[-1] = K.reshape(ds[-1], new_shape)
        # The output is a tensor.
        ds = K.concatenate(ds, -1)
    return ds


def _diag_gradients(ys, xs, order=1):
    """Returns the gradients of y in `ys` w.r.t. x in `xs`.

    `ys` and `xs` are each a Tensor or a list of tensors.

    # Arguments
        ys: A tensor or list of tesnors to be differentiated.
        xs: A tensor or list of tensors to be used for differentiation.
        order: Order of differentiation.

    # Returns
        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
    """
    assert ys.shape.as_list() == xs.shape.as_list(), \
        'Supported when X and Y has the same dimensions - ' + \
        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())

    ds = _gradients(ys, xs, order)
    return tf.linalg.diag_part(ds)


def _diag_gradients2(ys, xs, order=1):
    """Returns the gradients of y in `ys` w.r.t. x in `xs`.

    `ys` and `xs` are each a Tensor or a list of tensors.

    # Arguments
        ys: A tensor or list of tesnors to be differentiated.
        xs: A tensor or list of tensors to be used for differentiation.
        order: Order of differentiation.

    # Returns
        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
    """
    assert ys.shape.as_list() == xs.shape.as_list(), \
        'Supported when X and Y has the same dimensions - ' + \
        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())

    splitted_ys = tf.split(ys, num_or_size_splits=ys.shape[-1], axis=-1)
    ds = []
    for j, y in enumerate(splitted_ys):
        ds.append(y)
        for i in range(order):
            ds[-1] = unpack_singleton(
                tf.gradients(
                    ds[-1], xs,
                    unconnected_gradients='zero',
                    # colocate_gradients_with_ops=True, TF: V1.14.0
                )
            )
            
        ds[-1] = ds[-1][:, j:j+1]
    # The output is a tensor.
    ds = K.concatenate(ds, -1)
    return ds


def _div_gradients(ys, xs, order=1):
    """Returns the gradients of y in `ys` w.r.t. x in `xs`.

    `ys` and `xs` are each a Tensor or a list of tensors.

    # Arguments
        ys: A tensor or list of tesnors to be differentiated.
        xs: A tensor or list of tensors to be used for differentiation.
        order: Order of differentiation.

    # Returns
        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
    """
    assert ys.shape.as_list() == xs.shape.as_list(), \
        'Supported when X and Y has the same dimensions - ' + \
        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())

    ds = _diag_gradients(ys, xs, order)
    return tf.math.reduce_sum(ds, [1], keepdims=True)


def _lambda_gradient(ys, xs, order=1, gtype='Grad', name=''):
    """Returns the gradients of y in `ys` w.r.t. x in `xs` using Lambda layers.
    
    `ys` and `xs` are each a Tensor or a list of tensors.

    # Arguments
        ys: A tensor or list of tesnors to be differentiated.
        xs: A tensor or list of tensors to be used for differentiation.
        gtype: type of differentiation - can be:
            - 'Grad' for gradient operation, i.e. Gij = dy_j / dx_i
            - 'dGrad' for the diagonal of gradient tensor, i.e. Gi = dy_i / dx_i
            - 'Div' for divergence operation, i.e. G = sum(dy_i / dx_i)
        name: A str name for the Lambda layer. 

    # Returns
        A tuple, `(layers, grads)`.
        layers: A Lambda layer or list of Lambda layers where the gradient operator is applied.
        grads: A gradient tensor or list of gradient tensors. 
    """
    
    grads, layers = [], []
    for y in to_list(ys):
        if gtype == 'Grad':
            name_prefix = 'Grad_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
            lay = Lambda(lambda y: _gradients(y, xs, order), name=graph_unique_name(name_prefix))
        elif gtype == 'dGrad':
            name_prefix = 'DiagGrad_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
            lay = Lambda(lambda y: _diag_gradients(y, xs, order), name=graph_unique_name(name_prefix))
        elif gtype == 'Div':
            name_prefix = 'Div_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
            lay = Lambda(lambda y: _div_gradients(y, xs, order), name=graph_unique_name(name_prefix))
        else:
            raise ValueError(
                'Unrecognised gradient type: {} \n'.format(type) +
                '     Please choose among (Grad, dGrad, Div). '
            )
        layers += to_list(lay)
        grads += to_list(lay(y))

    return (unpack_singleton(layers), unpack_singleton(grads))


def _gdiff(gtype, f, *args, **kwargs):
    """Computes gradient of functional object f.

    # Arguments
        gtype: gradient type - choose from (Grad, dGrad, Div)
        f: Functional object.
        ys: layer name for `ys` to differentiate.
        xs: layer name for `xs` to be differentiated w.r.t.
        order: order of differentiation w.r.t. xs - defaulted to 1.

    # Returns
        A new functional object.
    """

    assert is_functional(f), \
        'Please provide a proper functional object. '
    assert(len(args) in (1,2)), \
        'Expected (`ys`, `xs`) or `xs` as the input, '\
        'was provided {:d} inputs'.format(len(args))
    if not all([isinstance(v, str) or is_functional(v) for v in args]):
        raise ValueError(
            'Expected a `Layer` name for a `Functional` to perform differentitation.'
        )

    try:
        inputs = f.inputs.copy()
        if len(args) == 0:
            x = unpack_singleton(f.inputs)
            assert is_tensor(x), \
                'multiple inputs detected - please provide an `x` name. '
            x_name = x.name.split('/')[0]
        else:
            x_id = 0 if len(args)==1 else 1
            if isinstance(args[x_id], str):
                x_name = args[x_id]
                x = next(l for l in f.layers if l.name == x_name).output
            elif is_functional(args[x_id]):
                inputs += to_list(args[x_id].inputs)
                x_lay = args[x_id].layers[-1]
                x_name = x_lay.name
                x = x_lay.output
            else:
                raise TypeError('Unsupported `x` entry. ')
            
        if len(args) <= 1:
            y = unpack_singleton(f.outputs)
            assert is_tensor(y), \
                'multiple outputs detected - please provide a `y` name. '
            y_name = y.name.split('/')[0]
        else:
            y_id = 0
            if isinstance(args[y_id], str):
                y_name = args[y_id]
                y = next(l for l in f.layers if l.name == y_name).output
            elif is_functional(args[y_id]):
                y_lay = args[y_id].layers[-1]
                y_name = y_lay.name
                y = y_lay.output
            else:
                raise TypeError('Unsupported `y` entry. ')
        
    except (StopIteration, ValueError):
        print("Did not find the layer {}. ".format(args))

    # check order of differentiation.
    order = 1
    if 'order' in kwargs.keys():
        order = kwargs['order']
    
    lay, tens = _lambda_gradient(
        y, x, order, gtype, "{}_{}".format(y_name, x_name)
    )
    
    Functional = type(f)
    res = Functional(
        inputs = unique_tensors(inputs),
        outputs = to_list(tens),
        layers = to_list(lay)
    )
    
    return res


def grad(f, *args, **kwargs):
    """Computes gradient tensor of functional object f.

    # Arguments
        f: Functional object.
        ys: layer name for `ys` to differentiate.
        xs: layer name for `xs` to be differentiated w.r.t.
        order: order of differentiation w.r.t. xs - defaulted to 1.

    # Returns
        A new functional object.
    """
    return _gdiff("Grad", f, *args, **kwargs)


# overlaod for backward compatibility 
diff = grad


# consistency with older versions.
def diag_grad(f, *args, **kwargs):
    """Computes diag of gradient tensor of functional object f.

    # Arguments
        f: Functional object.
        ys: layer name for `ys` to differentiate.
        xs: layer name for `xs` to be differentiated w.r.t.
        order: order of differentiation w.r.t. xs - defaulted to 1.

    # Returns
        A new functional object.
    """
    return _gdiff("dGrad", f, *args, **kwargs)


# consistency with older versions.
def div_grad(f, *args, **kwargs):
    """Computes Divergence of functional object f.

    # Arguments
        f: Functional object.
        ys: layer name for `ys` to differentiate.
        xs: layer name for `xs` to be differentiated w.r.t.
        order: order of differentiation w.r.t. xs - defaulted to 1.

    # Returns
        A new functional object.
    """
    return _gdiff("Div", f, *args, **kwargs)


def stop_gradient(f):
    """Equivalent to stop_grdient in tf.
    Use-case: to force zero-gradient with respect to a part of the network.

    # Arguments
        f: Functional object.

    # Returns
        A new functional object.
    """
    if is_functional(f):
        lmbd = [Lambda(lambda x: tf_stop_gradient(x), name=graph_unique_name("stop_grad")) for X in f.outputs]
    else:
        raise ValueError('Expected one functional in the arguments.')

    Functional = f.get_class()
    res = Functional(
        inputs = unique_tensors(f.inputs.copy()),
        outputs = _apply_operation(lmbd, f),
        layers = lmbd
    )

    return res


def radial_basis(xs, ci, radii):
    """Apply `radial_basis` function to x element-wise.

    # Arguments
        xs: List of functional objects.
        ci: Center of basis functional (same length as xs).
        radii: standard deviation or radius from the center.

    # Returns
        A new functional object.
    """
    assert len(xs) == len(ci)
    assert radii > 0.0
    assert all([is_variable(x) for x in xs])
    assert isinstance(xs, list) and isinstance(ci, list)

    for x in xs:
        validate_variable(x)

    return exp(-sum([(x - c)**2 for x, c in zip(xs, ci)])/radii**2)


def radial_basis2(xs, ci, radii):
    """Apply `radial_basis` function to x element-wise.

    # Arguments
        xs: List of functional objects.
        ci: Center of basis functional (same length as xs).
        radii: standard deviation or radius from the center.

    # Returns
        A new functional object.
    """
    assert len(xs) == len(ci)
    assert radii > 0.0
    assert all([is_variable(x) for x in xs])
    assert isinstance(xs, list) and isinstance(ci, list)

    for x in xs:
        validate_variable(x)

    d = xs[0] - ci[0]
    for i in range(1, len(xs)):
        d += xs[i] - ci[i]
    d /= radii
    
    return exp(-sum([(x - c)**2 for x, c in zip(xs, ci)])/radii**2)


def _warn_for_ndarray(other):
    if isinstance(other, np.ndarray):
        Warning(
            'Expecting `Tensor` objects instead of `ndarray`: ' +
            'Note data should go to the training process and ' +
            'this operation may break batch training. '
        )




""" Constraint class to condition on the training.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..utils import is_tensor


class Constraint(object):
    """ Configures the condition to impose constraint.

    # Arguments
        var: The layer name to impose the constraint.
        cond: A callable handle to function that imposes the condition.
        name: A `str` to be associated to the constraint.

    # Returns

    # Raises
        ValueError if `model` is not of class `SciModel`.
        ValueError for unrecognized inputs.
    """
    def __init__(self,
                 var=None,
                 cond=None,
                 name=None,
                 **kwargs):
        # check the inputs.
        assert isinstance(var, (str, type(None))), \
            "Expected a Layer Name of type str. "
        assert (cond is None or callable(cond) or is_tensor(cond)), \
            "Expected a function or a Tensor as input. "
        if 'ids' in kwargs or 'sol' in kwargs:
            raise ValueError(
                'Legacy interface: please use `SciModel` and `SciModel.train` to impose on specific ids. '
            )
        elif len(kwargs)>0:
            raise ValueError(
                'Unrecognized input variable: {}'.format(kwargs.keys())
            )
        assert isinstance(name, (str, type(None))), \
            "Expected a string input for the name. "
        self.var = var
        self.cond = cond
        self.name = name

    def __call__(self):
        return self.cond

    def eval(self, xs):
        return self.cond.eval(xs)
